use rand::Rng;
use std::time::Instant;

use crate::{OutfitSet, Projectile};
use galactica_content as content;

#[derive(Debug)]
pub struct Ship {
	pub handle: content::ShipHandle,
	pub faction: content::FactionHandle,
	pub outfits: OutfitSet,
	pub last_hit: Instant,

	// TODO: unified ship stats struct, like space
	// TODO: unified outfit stats struct: check
	pub hull: f32,
	pub shields: f32,
}

impl Ship {
	pub fn new(
		ct: &content::Content,
		handle: content::ShipHandle,
		faction: content::FactionHandle,
		outfits: OutfitSet,
	) -> Self {
		let s = ct.get_ship(handle);
		let shields = outfits.stat_sum().shield_strength;
		Ship {
			handle: handle,
			faction,
			outfits,
			hull: s.hull,
			shields,
			last_hit: Instant::now(),
		}
	}

	pub fn is_dead(&self) -> bool {
		self.hull <= 0.0
	}

	/// Called whenever a projectile hits this ship.
	/// Returns true if that projectile should be destroyed and false otherwise.
	pub fn handle_projectile_collision(&mut self, ct: &content::Content, p: &Projectile) -> bool {
		let f = ct.get_faction(self.faction);
		let r = f.relationships.get(&p.faction).unwrap();
		match r {
			content::Relationship::Hostile => {
				let mut d = p.content.damage;
				if self.is_dead() {
					return true;
				}
				if self.shields >= d {
					self.shields -= d
				} else {
					d -= self.shields;
					self.shields = 0.0;
					self.hull -= d;
				}
				self.last_hit = Instant::now();

				return true;
			}
			_ => return false,
		}
	}

	pub fn fire_guns(&mut self) -> Vec<(Projectile, content::GunPoint)> {
		if self.is_dead() {
			return vec![];
		}

		self.outfits
			.iter_guns_points()
			.filter(|(g, _)| g.cooldown <= 0.0)
			.map(|(g, p)| {
				let mut rng = rand::thread_rng();
				g.cooldown = g.kind.rate + &rng.gen_range(-g.kind.rate_rng..=g.kind.rate_rng);

				let lifetime = g.kind.projectile.lifetime
					+ rng.gen_range(
						-g.kind.projectile.lifetime_rng..=g.kind.projectile.lifetime_rng,
					);

				(
					Projectile {
						content: g.kind.projectile.clone(),
						lifetime: 0f32.max(lifetime),
						faction: self.faction,
					},
					p.clone(),
				)
			})
			.collect()
	}

	pub fn step(&mut self, t: f32) {
		let time_since = self.last_hit.elapsed().as_secs_f32();
		if self.shields != self.outfits.stat_sum().shield_strength {
			for (_, d, g) in &self.outfits.stat_sum().shield_generators {
				if time_since >= *d {
					self.shields += g * t;
					if self.shields > self.outfits.stat_sum().shield_strength {
						self.shields = self.outfits.stat_sum().shield_strength;
						break;
					}
				}
			}
		}
	}
}
